#define _XOPEN_SOURCE 500
#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <unistd.h>
#include <math.h>
#include <mpi.h>
#include <omp.h>

#include <tlhash.h>

#define MAX(a,b) ((a)>(b)?(a):(b))
#define MIN(a,b) ((a)<(b)?(a):(b))
#define pow2(x) ((x)*(x))

#define sched(x)// schedule(static, (x))

typedef int64_t int_t;
typedef double real_t;

typedef struct {
	int_t id;
    real_t x;
	real_t y;
	real_t z;
} vertex_t;

typedef struct t {
	int_t id;
	int_t id_local;
    int_t v[3];						// Indices of the 3 triangle vertices
    real_t c[3];					// Coordinates (x,y,z) of triangle centroid
    int_t n_neighbors;				// Neighbor count - 3 or fewer
    int_t neighbors_id[3];			// Element indices of up to 3 neighbors
    int_t neighbors_id_local[3];
	real_t area;					// Area of the triangle
} triangle_t;

typedef struct {
    int_t n_vertices_global, n_elements_global;
    int_t n_vertices_local, n_elements_local;
    int_t n_elements_local_max;
	int_t n_neumann_edges;
	int_t n_elements_local_with_halo;
	int_t n_elements_local_total;
	int_t n_upper_border, n_lower_border;
	int_t n_upper_halo, n_lower_halo;
	int_t upper_border_index, lower_border_index;
	int_t upper_halo_index, lower_halo_index;
	int_t neumann_edges_index;
    vertex_t *vertices;
    triangle_t *elements;
} mesh_t;


char* filename;

int
	size,
	rank;

int_t
	loop_block_neu,
	loop_block_tot,
	loop_block,
	warm_up_iter = 1000,
	time_samples = 40,
	time_count = 0,
    max_iter = 5001,
    snap_freq = 50;

real_t
	*calc_time,
	*halo_time,
	avg_calc = 0.0,
	avg_calc_p1 = 0.0,
	avg_calc_p2 = 0.0,
	avg_calc_p3 = 0.0,
	avg_halo = 0.0,
	variance_calc = 0.0,
	variance_calc_p1 = 0.0,
	variance_calc_p2 = 0.0,
	variance_calc_p3 = 0.0,
	variance_halo = 0.0;

mesh_t
    mesh;

real_t
    dt = 5e-2,
	average_weight = 0.4, //shouldn't be higher than 0.6
	*mass[2] = { NULL, NULL },
	*mass_velocity_x[2] = { NULL, NULL },
	*mass_velocity_y[2] = { NULL, NULL },
	*mass_velocity = NULL,
	*velocity_x = NULL,
	*velocity_y = NULL,
	*acceleration_x = NULL,
	*acceleration_y = NULL;

const real_t
	domain_size = 10.0,
	gravity = 9.81,
	density = 997.0;

real_t
	global_min_x = 0,
	global_min_y = 0,
	global_max_x = 0,
	global_max_y = 0;

#define PN(e) 			mass[0][(e)]
#define PN_next(e)		mass[1][(e)]
#define PNU(e)			mass_velocity_x[0][(e)]
#define PNU_next(e)		mass_velocity_x[1][(e)]
#define PNV(e)			mass_velocity_y[0][(e)]
#define PNV_next(e)		mass_velocity_y[1][(e)]
#define PNUV(e)			mass_velocity[(e)]
#define U(e)			velocity_x[(e)]
#define V(e)			velocity_y[(e)]
#define DU(e)			acceleration_x[(e)]
#define DV(e)			acceleration_y[(e)]
#define CALCT(e,x) 		calc_time[4*(e)+(x)]

void mesh_read ( mesh_t *mesh );
void mesh_connect ( mesh_t *mesh );
void mesh_finalize ( mesh_t *mesh );


void
init_state_smooth ( void )
{
    // Set initial mass as a function of distance from domain center
    for ( int_t idx=0; idx<mesh.n_elements_local_total; idx++ )
    {
        real_t
            y = mesh.elements[idx].c[1],
            x = mesh.elements[idx].c[0];
        real_t
            distance = sqrt (
                pow ( x-(global_max_x-global_min_x)/2, 2.0 ) + pow( y-(global_max_y-global_min_y)/2, 2.0 )
            );
        PN(idx) = exp ( -pow (distance, 4.0) );
		PNU(idx) = PNV(idx) = 0.0;
		PN(idx) += 0.3;
    }
}


void init_state_drop ( void )
{
    // Set initial mass as a function of distance from domain center
    for ( int_t idx=0; idx<mesh.n_elements_local_total; idx++ )
    {
        real_t
            cy = mesh.elements[idx].c[1] - global_max_y/2,
            cx = mesh.elements[idx].c[0] - global_max_x/2,
			global_distance_x = global_max_x - global_min_x,
			global_distance_y = global_max_y - global_min_y;
		PN(idx) = 1e-3;
		PNU(idx) = PNV(idx) = 0.0;
		if ( sqrt(cx*cx + cy*cy) < (global_distance_x + global_distance_y)/(2*10) )
			PN(idx) -= 1.5e-3 * exp (
				- 4*pow ( cx, 2.0 ) / global_max_x
				- 4*pow ( cy, 2.0 ) / global_max_y
			);
		PN(idx) *= density;
	}
	// Smooth out the init state
	for ( int_t iter=0; iter<1; iter++ )
	{
		for ( int_t idx=0; idx<mesh.n_elements_local; idx++ )
		{
			real_t pn_avg = PN(idx);
			real_t pnu_avg = PNU(idx);
			real_t pnv_avg = PNV(idx);
			triangle_t *element = &(mesh.elements[idx]);
			for ( int_t n=0; n<3; n++ )
			{
				int_t nidx = element->neighbors_id_local[n];
				pn_avg += PN(nidx);
				pnu_avg += PNU(nidx);
				pnv_avg += PNV(nidx);
			}
			PN(idx) = pn_avg / 4;
			PNU(idx) = pnu_avg / 4;
			PNV(idx) = pnv_avg / 4;
		}
	}
}


void
time_step ( void )
{
	// Boundary conditions
	#pragma omp parallel for sched(loop_block_neu)
	for ( int_t idx=mesh.neumann_edges_index; idx<mesh.n_elements_local_total; idx++ )
	{
		triangle_t *element = &(mesh.elements[idx]);
		int_t nidx = element->neighbors_id_local[0];
		PN(idx) = PN(nidx);
		PNU(idx) = -PNU(nidx);
		PNV(idx) = -PNV(nidx);
	}

	#pragma omp parallel for sched(loop_block_tot)
	for ( int_t idx=0; idx<mesh.n_elements_local_total; idx++ )
	{
		U(idx) = PNU(idx) / PN(idx);
		V(idx) = PNV(idx) / PN(idx);
	}

	#pragma omp parallel for sched(loop_block_tot)
	for ( int_t idx=0; idx<mesh.n_elements_local_total; idx++ )
	{
		PNUV(idx) = PN(idx) * U(idx) * V(idx);
	}

	#pragma omp parallel for sched(loop_block_tot)
	for ( int_t idx=0; idx<mesh.n_elements_local_total; idx++ )
	{
		DU(idx) = PN(idx) * U(idx) * U(idx)
			+ 0.5 * PN(idx) * (PN(idx) / density) * gravity;
		DV(idx) = PN(idx) * V(idx) * V(idx)
			+ 0.5 * PN(idx) * (PN(idx) / density) * gravity;
	}

	#pragma omp parallel for sched(loop_block)
	for ( int_t idx=0; idx<mesh.n_elements_local; idx++ )
	{
		real_t pnu_avg = 0.0;
		real_t pnv_avg = 0.0;
        real_t pnuv_grad = 0.0;
        real_t du_grad = 0.0;
        real_t dv_grad = 0.0;
		triangle_t *element = &(mesh.elements[idx]);
		for ( int_t n=0; n<3; n++ )
		{
			int_t nidx = element->neighbors_id_local[n];
			triangle_t *neighbor = &(mesh.elements[nidx]);
			real_t distance = sqrt (
				pow2(neighbor->c[0]-element->c[0])
				+ pow2(neighbor->c[1]-element->c[1])
			);
			pnu_avg += PNU(nidx);
			pnv_avg += PNV(nidx);
			pnuv_grad += (PNUV(nidx) - PNUV(idx)) / distance;
			du_grad += (DU(nidx) - DU(idx)) / distance;
			dv_grad += (DV(nidx) - DV(idx)) / distance;
		}
		pnu_avg *= (1 - average_weight) / 3;
		pnv_avg *= (1 - average_weight) / 3;
		pnu_avg += PNU(idx) * average_weight;
		pnv_avg += PNV(idx) * average_weight;
		pnuv_grad /= 3;
		du_grad /= 3;
		dv_grad /= 3;
		PNU_next(idx) = pnu_avg - dt * (du_grad + pnuv_grad);
		PNV_next(idx) = pnv_avg - dt * (dv_grad + pnuv_grad);
	}

	#pragma omp parallel for sched(loop_block)
	for ( int_t idx=0; idx<mesh.n_elements_local; idx++ )
	{
		real_t pn_avg = 0.0;
        real_t pnu_grad = 0.0;
        real_t pnv_grad = 0.0;
		triangle_t *element = &(mesh.elements[idx]);
		for ( int_t n=0; n<3; n++ )
		{
            int_t nidx = element->neighbors_id_local[n];
			triangle_t *neighbor = &(mesh.elements[nidx]);
            real_t distance = sqrt (
                pow2(neighbor->c[0]-element->c[0])
                + pow2(neighbor->c[1]-element->c[1])
            );
			pn_avg += PN(nidx);
			pnu_grad += (PNU(nidx) - PNU(idx)) / distance;
			pnv_grad += (PNV(nidx) - PNV(idx)) / distance;
		}
		pn_avg *= (1 - average_weight) / 3;
		pn_avg += PN(idx) * average_weight;
		pnu_grad /= 3;
		pnv_grad /= 3;
		PN_next(idx) = pn_avg - dt * (pnu_grad + pnv_grad);
	}
}


void
time_step_measure ( void )
{
	real_t
		t_start,
		t_end;
	
	t_start = MPI_Wtime();
	// Boundary conditions
	#pragma omp parallel for sched(loop_block_neu)
	for ( int_t idx=mesh.neumann_edges_index; idx<mesh.n_elements_local_total; idx++ )
	{
		triangle_t *element = &(mesh.elements[idx]);
		int_t nidx = element->neighbors_id_local[0];
		PN(idx) = PN(nidx);
		PNU(idx) = -PNU(nidx);
		PNV(idx) = -PNV(nidx);
	}

	#pragma omp parallel for sched(loop_block_tot)
	for ( int_t idx=0; idx<mesh.n_elements_local_total; idx++ )
	{
		U(idx) = PNU(idx) / PN(idx);
		V(idx) = PNV(idx) / PN(idx);
	}

	#pragma omp parallel for sched(loop_block_tot)
	for ( int_t idx=0; idx<mesh.n_elements_local_total; idx++ )
	{
		PNUV(idx) = PN(idx) * U(idx) * V(idx);
	}

	#pragma omp parallel for sched(loop_block_tot)
	for ( int_t idx=0; idx<mesh.n_elements_local_total; idx++ )
	{
		DU(idx) = PN(idx) * U(idx) * U(idx)
			+ 0.5 * PN(idx) * (PN(idx) / density) * gravity;
		DV(idx) = PN(idx) * V(idx) * V(idx)
			+ 0.5 * PN(idx) * (PN(idx) / density) * gravity;
	}
	t_end = MPI_Wtime();
	CALCT(time_count, 1) = t_end - t_start;

	t_start = MPI_Wtime();
	#pragma omp parallel for sched(loop_block)
	for ( int_t idx=0; idx<mesh.n_elements_local; idx++ )
	{
		real_t pnu_avg = 0.0;
		real_t pnv_avg = 0.0;
        real_t pnuv_grad = 0.0;
        real_t du_grad = 0.0;
        real_t dv_grad = 0.0;
		triangle_t *element = &(mesh.elements[idx]);
		for ( int_t n=0; n<3; n++ )
		{
			int_t nidx = element->neighbors_id_local[n];
			triangle_t *neighbor = &(mesh.elements[nidx]);
			real_t distance = sqrt (
				pow2(neighbor->c[0]-element->c[0])
				+ pow2(neighbor->c[1]-element->c[1])
			);
			pnu_avg += PNU(nidx);
			pnv_avg += PNV(nidx);
			pnuv_grad += (PNUV(nidx) - PNUV(idx)) / distance;
			du_grad += (DU(nidx) - DU(idx)) / distance;
			dv_grad += (DV(nidx) - DV(idx)) / distance;
		}
		pnu_avg *= (1 - average_weight) / 3;
		pnv_avg *= (1 - average_weight) / 3;
		pnu_avg += PNU(idx) * average_weight;
		pnv_avg += PNV(idx) * average_weight;
		pnuv_grad /= 3;
		du_grad /= 3;
		dv_grad /= 3;
		PNU_next(idx) = pnu_avg - dt * (du_grad + pnuv_grad);
		PNV_next(idx) = pnv_avg - dt * (dv_grad + pnuv_grad);
	}
	t_end = MPI_Wtime();
	CALCT(time_count, 2) = t_end - t_start;

	t_start = MPI_Wtime();
	#pragma omp parallel for sched(loop_block)
	for ( int_t idx=0; idx<mesh.n_elements_local; idx++ )
	{
		real_t pn_avg = 0.0;
        real_t pnu_grad = 0.0;
        real_t pnv_grad = 0.0;
		triangle_t *element = &(mesh.elements[idx]);
		for ( int_t n=0; n<3; n++ )
		{
            int_t nidx = element->neighbors_id_local[n];
			triangle_t *neighbor = &(mesh.elements[nidx]);
            real_t distance = sqrt (
                pow2(neighbor->c[0]-element->c[0])
                + pow2(neighbor->c[1]-element->c[1])
            );
			pn_avg += PN(nidx);
			pnu_grad += (PNU(nidx) - PNU(idx)) / distance;
			pnv_grad += (PNV(nidx) - PNV(idx)) / distance;
		}
		pn_avg *= (1 - average_weight) / 3;
		pn_avg += PN(idx) * average_weight;
		pnu_grad /= 3;
		pnv_grad /= 3;
		PN_next(idx) = pn_avg - dt * (pnu_grad + pnv_grad);
	}
	t_end = MPI_Wtime();
	CALCT(time_count, 3) = t_end - t_start;
}


void
halo_exchange( void )
{
	if (rank != 0) {
		int_t peer = rank - 1;
		MPI_Sendrecv(&PN(mesh.upper_border_index), mesh.n_upper_border, MPI_DOUBLE, peer, rank*4,
					 &PN(mesh.upper_halo_index), mesh.n_upper_halo, MPI_DOUBLE, peer, peer*4+1,
					 MPI_COMM_WORLD, MPI_STATUS_IGNORE
		);
		MPI_Sendrecv(&PNU(mesh.upper_border_index), mesh.n_upper_border, MPI_DOUBLE, peer, rank*4,
					 &PNU(mesh.upper_halo_index), mesh.n_upper_halo, MPI_DOUBLE, peer, peer*4+1,
					 MPI_COMM_WORLD, MPI_STATUS_IGNORE
		);
		MPI_Sendrecv(&PNV(mesh.upper_border_index), mesh.n_upper_border, MPI_DOUBLE, peer, rank*4,
					 &PNV(mesh.upper_halo_index), mesh.n_upper_halo, MPI_DOUBLE, peer, peer*4+1,
					 MPI_COMM_WORLD, MPI_STATUS_IGNORE
		);
	}
	if (rank != (size - 1)) {
		int_t peer = rank + 1;
		MPI_Sendrecv(&PN(mesh.lower_border_index), mesh.n_lower_border, MPI_DOUBLE, peer, rank*4+1,
					 &PN(mesh.lower_halo_index), mesh.n_lower_halo, MPI_DOUBLE, peer, peer*4,
					 MPI_COMM_WORLD, MPI_STATUS_IGNORE
		);
		MPI_Sendrecv(&PNU(mesh.lower_border_index), mesh.n_lower_border, MPI_DOUBLE, peer, rank*4+1,
					 &PNU(mesh.lower_halo_index), mesh.n_lower_halo, MPI_DOUBLE, peer, peer*4,
					 MPI_COMM_WORLD, MPI_STATUS_IGNORE
		);
		MPI_Sendrecv(&PNV(mesh.lower_border_index), mesh.n_lower_border, MPI_DOUBLE, peer, rank*4+1,
					 &PNV(mesh.lower_halo_index), mesh.n_lower_halo, MPI_DOUBLE, peer, peer*4,
					 MPI_COMM_WORLD, MPI_STATUS_IGNORE
		);
	}
}


void
halo_exchange_sendrecv( void )
{
	int_t peer_up = rank == 0 ? MPI_PROC_NULL : rank - 1;
	int_t peer_down = rank == (size - 1) ? MPI_PROC_NULL : rank + 1;
	int_t peer_up_tag = rank == 0 ? 0 : rank - 1;
	int_t peer_down_tag = rank == (size - 1) ? 0 : rank + 1;

	MPI_Sendrecv(&PN(mesh.upper_border_index), mesh.n_upper_border, MPI_DOUBLE, peer_up, rank,
				 &PN(mesh.lower_halo_index), mesh.n_lower_halo, MPI_DOUBLE, peer_down, peer_down_tag,
				 MPI_COMM_WORLD, MPI_STATUS_IGNORE
	);
	MPI_Sendrecv(&PNU(mesh.upper_border_index), mesh.n_upper_border, MPI_DOUBLE, peer_up, rank+(size*2),
				 &PNU(mesh.lower_halo_index), mesh.n_lower_halo, MPI_DOUBLE, peer_down, peer_down_tag+(size*2),
				 MPI_COMM_WORLD, MPI_STATUS_IGNORE
	);
	MPI_Sendrecv(&PNV(mesh.upper_border_index), mesh.n_upper_border, MPI_DOUBLE, peer_up, rank+(size*4),
				 &PNV(mesh.lower_halo_index), mesh.n_lower_halo, MPI_DOUBLE, peer_down, peer_down_tag+(size*4),
				 MPI_COMM_WORLD, MPI_STATUS_IGNORE
	);

	MPI_Sendrecv(&PN(mesh.lower_border_index), mesh.n_lower_border, MPI_DOUBLE, peer_down, rank+size,
				 &PN(mesh.upper_halo_index), mesh.n_upper_halo, MPI_DOUBLE, peer_up, peer_up_tag+size,
				 MPI_COMM_WORLD, MPI_STATUS_IGNORE
	);
	MPI_Sendrecv(&PNU(mesh.lower_border_index), mesh.n_lower_border, MPI_DOUBLE, peer_down, rank+(size*3),
				 &PNU(mesh.upper_halo_index), mesh.n_upper_halo, MPI_DOUBLE, peer_up, peer_up_tag+(size*3),
				 MPI_COMM_WORLD, MPI_STATUS_IGNORE
	);
	MPI_Sendrecv(&PNV(mesh.lower_border_index), mesh.n_lower_border, MPI_DOUBLE, peer_down, rank+(size*5),
				 &PNV(mesh.upper_halo_index), mesh.n_upper_halo, MPI_DOUBLE, peer_up, peer_up_tag+(size*5),
				 MPI_COMM_WORLD, MPI_STATUS_IGNORE
	);
}


void
halo_exchange_send_recv( void )
{
	int_t peer_up = rank - 1;
	int_t peer_down = rank + 1;
	if (rank != 0)
		MPI_Send(&PN(mesh.upper_border_index), mesh.n_upper_border, MPI_DOUBLE, peer_up, rank, MPI_COMM_WORLD);
	if (rank != (size - 1))
		MPI_Recv(&PN(mesh.lower_halo_index), mesh.n_lower_halo, MPI_DOUBLE, peer_down, peer_down, MPI_COMM_WORLD, MPI_STATUS_IGNORE);
	if (rank != 0)
		MPI_Send(&PNU(mesh.upper_border_index), mesh.n_upper_border, MPI_DOUBLE, peer_up, rank+(size*2), MPI_COMM_WORLD);
	if (rank != (size - 1))
		MPI_Recv(&PNU(mesh.lower_halo_index), mesh.n_lower_halo, MPI_DOUBLE, peer_down, peer_down+(size*2), MPI_COMM_WORLD, MPI_STATUS_IGNORE);
	if (rank != 0)
		MPI_Send(&PNV(mesh.upper_border_index), mesh.n_upper_border, MPI_DOUBLE, peer_up, rank+(size*4), MPI_COMM_WORLD);
	if (rank != (size - 1))
		MPI_Recv(&PNV(mesh.lower_halo_index), mesh.n_lower_halo, MPI_DOUBLE, peer_down, peer_down+(size*4), MPI_COMM_WORLD, MPI_STATUS_IGNORE);

	if (rank != (size - 1))
		MPI_Send(&PN(mesh.lower_border_index), mesh.n_lower_border, MPI_DOUBLE, peer_down, rank+size, MPI_COMM_WORLD);
	if (rank != 0)
		MPI_Recv(&PN(mesh.upper_halo_index), mesh.n_upper_halo, MPI_DOUBLE, peer_up, peer_up+size, MPI_COMM_WORLD, MPI_STATUS_IGNORE);
	if (rank != (size - 1))
		MPI_Send(&PNU(mesh.lower_border_index), mesh.n_lower_border, MPI_DOUBLE, peer_down, rank+(size*3), MPI_COMM_WORLD);
	if (rank != 0)
		MPI_Recv(&PNU(mesh.upper_halo_index), mesh.n_upper_halo, MPI_DOUBLE, peer_up, peer_up+(size*3), MPI_COMM_WORLD, MPI_STATUS_IGNORE);
	if (rank != (size - 1))
		MPI_Send(&PNV(mesh.lower_border_index), mesh.n_lower_border, MPI_DOUBLE, peer_down, rank+(size*5), MPI_COMM_WORLD);
	if (rank != 0)
		MPI_Recv(&PNV(mesh.upper_halo_index), mesh.n_upper_halo, MPI_DOUBLE, peer_up, peer_up+(size*5), MPI_COMM_WORLD, MPI_STATUS_IGNORE);
}


void
halo_exchange_non_block( void )
{
	MPI_Request reqs[12];
	MPI_Status stats[12];
	int n_reqs = 0;

	if (rank != 0) {
		int_t peer = rank - 1;
		MPI_Isend(&PN(mesh.upper_border_index), mesh.n_upper_border, MPI_DOUBLE, peer, rank,
					 MPI_COMM_WORLD, &reqs[n_reqs++]
		);
		MPI_Isend(&PNU(mesh.upper_border_index), mesh.n_upper_border, MPI_DOUBLE, peer, rank+(size*2),
					 MPI_COMM_WORLD, &reqs[n_reqs++]
		);
		MPI_Isend(&PNV(mesh.upper_border_index), mesh.n_upper_border, MPI_DOUBLE, peer, rank+(size*4),
					 MPI_COMM_WORLD, &reqs[n_reqs++]
		);
	}
	if (rank != (size - 1)) {
		int_t peer = rank + 1;
		MPI_Isend(&PN(mesh.lower_border_index), mesh.n_lower_border, MPI_DOUBLE, peer, rank+size,
					 MPI_COMM_WORLD, &reqs[n_reqs++]
		);
		MPI_Isend(&PNU(mesh.lower_border_index), mesh.n_lower_border, MPI_DOUBLE, peer, rank+(size*3),
					 MPI_COMM_WORLD, &reqs[n_reqs++]
		);
		MPI_Isend(&PNV(mesh.lower_border_index), mesh.n_lower_border, MPI_DOUBLE, peer, rank+(size*5),
					 MPI_COMM_WORLD, &reqs[n_reqs++]
		);
	}
	if (rank != 0) {
		int_t peer = rank - 1;
		MPI_Irecv(&PN(mesh.upper_halo_index), mesh.n_upper_halo, MPI_DOUBLE, peer, peer+size,
					 MPI_COMM_WORLD, &reqs[n_reqs++]
		);
		MPI_Irecv(&PNU(mesh.upper_halo_index), mesh.n_upper_halo, MPI_DOUBLE, peer, peer+(size*3),
					 MPI_COMM_WORLD, &reqs[n_reqs++]
		);
		MPI_Irecv(&PNV(mesh.upper_halo_index), mesh.n_upper_halo, MPI_DOUBLE, peer, peer+(size*5),
					 MPI_COMM_WORLD, &reqs[n_reqs++]
		);
	}
	if (rank != (size - 1)) {
		int_t peer = rank + 1;
		MPI_Irecv(&PN(mesh.lower_halo_index), mesh.n_lower_halo, MPI_DOUBLE, peer, peer,
					 MPI_COMM_WORLD, &reqs[n_reqs++]
		);
		MPI_Irecv(&PNU(mesh.lower_halo_index), mesh.n_lower_halo, MPI_DOUBLE, peer, peer+(size*2),
					 MPI_COMM_WORLD, &reqs[n_reqs++]
		);
		MPI_Irecv(&PNV(mesh.lower_halo_index), mesh.n_lower_halo, MPI_DOUBLE, peer, peer+(size*4),
					 MPI_COMM_WORLD, &reqs[n_reqs++]
		);
	}
	MPI_Waitall(n_reqs, reqs, stats);
}


void
save_state ( int_t index )
{
    char filename[256];
    memset ( filename, 0, 256 );
    sprintf ( filename, "data/%.5ld.dat", index );

	MPI_File fh;
	int access_mode = MPI_MODE_CREATE | MPI_MODE_WRONLY | MPI_MODE_APPEND;
	if (MPI_File_open(MPI_COMM_WORLD, filename, access_mode, MPI_INFO_NULL, &fh) != MPI_SUCCESS) {
		printf("Rank %d failed to open file\n", rank);
		MPI_Abort(MPI_COMM_WORLD, EXIT_FAILURE);
	}
    for ( int_t idx=0; idx<mesh.n_elements_local_max; idx++ ) {
		vertex_t c = { mesh.elements[idx].c[0], mesh.elements[idx].c[1], mesh.elements[idx].c[2] };
		if (idx < mesh.n_elements_local) {
			int size = snprintf(NULL, 0, "%lf %lf %lf\n",
								mesh.elements[idx].c[0],
								mesh.elements[idx].c[1],
								mass[0][idx]
			);
			char *out = malloc(size + 1);
			sprintf ( out, "%lf %lf %lf\n",
					  mesh.elements[idx].c[0],
					  mesh.elements[idx].c[1],
					  mass[0][idx]
			);
			MPI_File_write_ordered(fh, out, size, MPI_CHAR, MPI_STATUS_IGNORE);
			free(out);
		} else {
			MPI_File_write_ordered(fh, NULL, 0, MPI_CHAR, MPI_STATUS_IGNORE);
		}
	}
	if (MPI_File_close(&fh) != MPI_SUCCESS) {
		printf("Rank %d failed to close file\n", rank);
		MPI_Abort(MPI_COMM_WORLD, EXIT_FAILURE);
	}
}


int
main ( int argc, char **argv )
{
	MPI_Init(&argc, &argv);
	MPI_Comm_size(MPI_COMM_WORLD, &size);
	MPI_Comm_rank(MPI_COMM_WORLD, &rank);

    filename = argv[1];
    sprintf ( filename, "%s/%d.txt", filename, rank );

	mesh_read ( &mesh );

	mass[0] = malloc ( mesh.n_elements_local_total * sizeof(real_t) );
	mass[1] = malloc ( mesh.n_elements_local_total * sizeof(real_t) );
	mass_velocity_x[0] = malloc ( mesh.n_elements_local_total * sizeof(real_t) );
	mass_velocity_x[1] = malloc ( mesh.n_elements_local_total * sizeof(real_t) );
	mass_velocity_y[0] = malloc ( mesh.n_elements_local_total * sizeof(real_t) );
	mass_velocity_y[1] = malloc ( mesh.n_elements_local_total * sizeof(real_t) );
	mass_velocity = malloc ( mesh.n_elements_local_total * sizeof(real_t) );
	velocity_x = malloc ( mesh.n_elements_local_total * sizeof(real_t) );
	velocity_y = malloc ( mesh.n_elements_local_total * sizeof(real_t) );
	acceleration_x = malloc ( mesh.n_elements_local_total * sizeof(real_t) );
	acceleration_y = malloc ( mesh.n_elements_local_total * sizeof(real_t) );

	calc_time = malloc ( 4 * time_samples * sizeof(real_t) );
	halo_time = malloc ( time_samples * sizeof(real_t) );
	real_t t_calc_start;
	real_t t_calc_end;
	real_t t_halo_start;
	real_t t_halo_end;
	int_t time_freq = max_iter / time_samples;
	if (max_iter%time_samples)
		time_freq++;

	loop_block_neu = mesh.n_neumann_edges / omp_get_max_threads();
	loop_block_tot = mesh.n_elements_local_total / omp_get_max_threads();
	loop_block = mesh.n_elements_local / omp_get_max_threads();

	mesh_connect( &mesh );

	for ( int_t iter=0; iter<warm_up_iter; iter++ )
	{
		time_step();
	}

    init_state_drop();

    for ( int_t iter=0; iter<max_iter; iter++ )
    {
		if ((iter%time_freq) == 0) {
			MPI_Barrier(MPI_COMM_WORLD);
			t_halo_start = MPI_Wtime();
			halo_exchange_non_block();
			t_halo_end = MPI_Wtime();
			halo_time[time_count] = t_halo_end - t_halo_start;
		
			/*if ( (iter%snap_freq)==0 )
			{
				if (rank == 0)
					printf ( "%ld / %ld\n", iter, max_iter );
				save_state ( iter / snap_freq );
			}*/

			MPI_Barrier(MPI_COMM_WORLD);
			t_calc_start = MPI_Wtime();
			time_step();
			t_calc_end = MPI_Wtime();
			CALCT(time_count, 0) = t_calc_end - t_calc_start;

			time_step_measure();

			time_count++;
		} else {
			MPI_Barrier(MPI_COMM_WORLD);
			halo_exchange_non_block();
			
			/*if ( (iter%snap_freq)==0 )
			{
				if (rank == 0)
					printf ( "%ld / %ld\n", iter, max_iter );
				save_state ( iter / snap_freq );
			}*/

			time_step();
		}

        real_t *tmp = mass[0];
        mass[0] = mass[1];
        mass[1] = tmp;
        tmp = mass_velocity_x[0];
        mass_velocity_x[0] = mass_velocity_x[1];
        mass_velocity_x[1] = tmp;
        tmp = mass_velocity_y[0];
        mass_velocity_y[0] = mass_velocity_y[1];
        mass_velocity_y[1] = tmp;
	}

	// Calculate time
	MPI_Barrier(MPI_COMM_WORLD);
	if (rank == 0) {
		printf("FVM Triangle\n");
		printf("%d (number of processes)\n", size);
		printf("%d (number of threads)\n", omp_get_max_threads());
		printf("%ld (number of vertices)\n", mesh.n_vertices_global);
		printf("%ld (number of elements)\n", mesh.n_elements_global);
		printf("%ld (number of samples)\n\n", time_samples);
	}
	MPI_Barrier(MPI_COMM_WORLD);
	for (int_t i=0; i<time_samples; i++) {
		avg_calc += CALCT(i, 0);
		avg_calc_p1 += CALCT(i, 1);
		avg_calc_p2 += CALCT(i, 2);
		avg_calc_p3 += CALCT(i, 3);
		avg_halo += halo_time[i];
	}
	avg_calc /= time_samples;
	avg_calc_p1 /= time_samples;
	avg_calc_p2 /= time_samples;
	avg_calc_p3 /= time_samples;
	avg_halo /= time_samples;
	for (int_t i=0; i<time_samples; i++) {
		variance_calc += pow2( CALCT(i, 0) - avg_calc );
		variance_calc_p1 += pow2( CALCT(i, 1) - avg_calc );
		variance_calc_p2 += pow2( CALCT(i, 2) - avg_calc );
		variance_calc_p3 += pow2( CALCT(i, 3) - avg_calc );
		variance_halo += pow2( halo_time[i] - avg_halo );
	}
	variance_calc /= (time_samples - 1);
	variance_calc_p1 /= (time_samples - 1);
	variance_calc_p2 /= (time_samples - 1);
	variance_calc_p3 /= (time_samples - 1);
	variance_halo /= (time_samples - 1);
	variance_calc = sqrt( variance_calc );
	variance_calc_p1 = sqrt( variance_calc_p1 );
	variance_calc_p2 = sqrt( variance_calc_p2 );
	variance_calc_p3 = sqrt( variance_calc_p3 );
	variance_halo = sqrt( variance_halo );

	MPI_Barrier(MPI_COMM_WORLD);
	real_t
		com_avg_calc = avg_calc,
		com_avg_calc_p1 = avg_calc_p1,
		com_avg_calc_p2 = avg_calc_p2,
		com_avg_calc_p3 = avg_calc_p3,
		com_avg_halo = avg_halo,
		com_variance_calc = variance_calc,
		com_variance_calc_p1 = variance_calc_p1,
		com_variance_calc_p2 = variance_calc_p2,
		com_variance_calc_p3 = variance_calc_p3,
		com_variance_halo = variance_halo;
	for (int_t i=0; i<size; i++) {
		if (i == rank && i != 0) {
			MPI_Send(&avg_calc, 1, MPI_DOUBLE, 0, 0, MPI_COMM_WORLD);
			MPI_Send(&avg_calc_p1, 1, MPI_DOUBLE, 0, 1, MPI_COMM_WORLD);
			MPI_Send(&avg_calc_p2, 1, MPI_DOUBLE, 0, 2, MPI_COMM_WORLD);
			MPI_Send(&avg_calc_p3, 1, MPI_DOUBLE, 0, 3, MPI_COMM_WORLD);
			MPI_Send(&avg_halo, 1, MPI_DOUBLE, 0, 4, MPI_COMM_WORLD);
			MPI_Send(&variance_calc, 1, MPI_DOUBLE, 0, 5, MPI_COMM_WORLD);
			MPI_Send(&variance_calc_p1, 1, MPI_DOUBLE, 0, 6, MPI_COMM_WORLD);
			MPI_Send(&variance_calc_p2, 1, MPI_DOUBLE, 0, 7, MPI_COMM_WORLD);
			MPI_Send(&variance_calc_p3, 1, MPI_DOUBLE, 0, 8, MPI_COMM_WORLD);
			MPI_Send(&variance_halo, 1, MPI_DOUBLE, 0, 9, MPI_COMM_WORLD);
		}
		if (rank == 0 && i != 0) {
			MPI_Recv(&com_avg_calc, 1, MPI_DOUBLE, i, 0, MPI_COMM_WORLD, MPI_STATUS_IGNORE);
			MPI_Recv(&com_avg_calc_p1, 1, MPI_DOUBLE, i, 1, MPI_COMM_WORLD, MPI_STATUS_IGNORE);
			MPI_Recv(&com_avg_calc_p2, 1, MPI_DOUBLE, i, 2, MPI_COMM_WORLD, MPI_STATUS_IGNORE);
			MPI_Recv(&com_avg_calc_p3, 1, MPI_DOUBLE, i, 3, MPI_COMM_WORLD, MPI_STATUS_IGNORE);
			MPI_Recv(&com_avg_halo, 1, MPI_DOUBLE, i, 4, MPI_COMM_WORLD, MPI_STATUS_IGNORE);
			MPI_Recv(&com_variance_calc, 1, MPI_DOUBLE, i, 5, MPI_COMM_WORLD, MPI_STATUS_IGNORE);
			MPI_Recv(&com_variance_calc_p1, 1, MPI_DOUBLE, i, 6, MPI_COMM_WORLD, MPI_STATUS_IGNORE);
			MPI_Recv(&com_variance_calc_p2, 1, MPI_DOUBLE, i, 7, MPI_COMM_WORLD, MPI_STATUS_IGNORE);
			MPI_Recv(&com_variance_calc_p3, 1, MPI_DOUBLE, i, 8, MPI_COMM_WORLD, MPI_STATUS_IGNORE);
			MPI_Recv(&com_variance_halo, 1, MPI_DOUBLE, i, 9, MPI_COMM_WORLD, MPI_STATUS_IGNORE);
		}
		MPI_Barrier(MPI_COMM_WORLD);
		if (rank == 0) {
			printf("RANK %ld\n"
					"%lf +- %lf ms (calc avg)\n"
					"%lf +- %lf ms (subcalc 1 avg)\n"
					"%lf +- %lf ms (subcalc 2 avg)\n"
					"%lf +- %lf ms (subcalc 3 avg)\n"
					"%lf +- %lf ms (border exchange avg)\n\n",
					i,
					com_avg_calc*1000, com_variance_calc*1000,
					com_avg_calc_p1*1000, com_variance_calc_p1*1000,
					com_avg_calc_p2*1000, com_variance_calc_p2*1000,
					com_avg_calc_p3*1000, com_variance_calc_p3*1000,
					com_avg_halo*1000, com_variance_halo*1000
			);
		}
		MPI_Barrier(MPI_COMM_WORLD);
	}

    mesh_finalize ( &mesh );
	free ( mass[0] );
	free ( mass[1] );
	free ( mass_velocity_x[0] );
	free ( mass_velocity_x[1] );
	free ( mass_velocity_y[0] );
	free ( mass_velocity_y[1] );
	free ( mass_velocity );
	free ( velocity_x );
	free ( velocity_y );
	free ( acceleration_x );
	free ( acceleration_y );
	free ( calc_time );
	free ( halo_time );

	MPI_Finalize();

    exit ( EXIT_SUCCESS );
}

/* Fix data structures
 */
void
mesh_connect ( mesh_t *mesh )
{
	#pragma omp parallel for sched(loop_block_tot)
	for (int_t i=0; i<mesh->n_elements_local_total; i++) {
		triangle_t *element = &(mesh->elements[i]);
		for (int_t n=0; n<3; n++) {
			int_t nidx = element->neighbors_id[n];
			if (nidx >= 0) {
				for (int_t j=0; j<mesh->n_elements_local_total; j++) {
					int_t idx = mesh->elements[j].id;
					if (idx == nidx) {
						element->neighbors_id_local[n] = j;
						break;
					}
				}
			}
		}
	}
}

/* Read triangle mesh from python script.
 */
void
mesh_read ( mesh_t *mesh )
{
	FILE *fp = fopen(filename, "r");
	if (fp == NULL)
		exit(EXIT_FAILURE);

    int err = fscanf ( fp,
        "%ld %ld %lf %lf %lf %lf\n", &(mesh->n_vertices_global), &(mesh->n_elements_global),
									 &(global_min_x), &(global_min_y), &(global_max_x), &(global_max_y)
    );
    if ( err != 6 )
    {
        fprintf ( stderr, "Error: unable to read from file\n" );
        exit ( EXIT_FAILURE );
    }
	err = fscanf (fp,
		"%ld %ld %ld %ld\n", &(mesh->n_vertices_local), &(mesh->n_elements_local), &(mesh->n_elements_local_max), &(mesh->n_neumann_edges)
	);
	if ( err != 4 )
    {
        fprintf ( stderr, "Error: unable to read from file\n" );
        exit ( EXIT_FAILURE );
    }
	err = fscanf ( fp,
		"%ld %ld %ld %ld\n", &(mesh->n_upper_border), &(mesh->n_lower_border), &(mesh->n_upper_halo), &(mesh->n_lower_halo)
	);
	if ( err != 4 )
    {
        fprintf ( stderr, "Error: unable to read from file\n" );
        exit ( EXIT_FAILURE );
    }

	mesh->n_elements_local_with_halo = mesh->n_elements_local + mesh->n_upper_halo + mesh->n_lower_halo;
	mesh->n_elements_local_total = mesh->n_elements_local_with_halo + mesh->n_neumann_edges;
	mesh->upper_border_index = mesh->n_elements_local - mesh->n_lower_border - mesh->n_upper_border;
	mesh->lower_border_index = mesh->n_elements_local - mesh->n_lower_border;
	mesh->upper_halo_index = mesh->n_elements_local;
	mesh->lower_halo_index = mesh->n_elements_local + mesh->n_upper_halo;
	mesh->neumann_edges_index = mesh->n_elements_local_with_halo;
    mesh->vertices = malloc ( mesh->n_vertices_local * sizeof(vertex_t) );
    mesh->elements = malloc ( mesh->n_elements_local_total * sizeof(triangle_t) );

	/*printf("rank %d, %ld %ld %ld - %ld %ld %ld %ld %ld - %ld %ld %ld %ld %ld\n", rank,
			mesh->n_elements_local, mesh->n_elements_local_with_halo, mesh->n_elements_local_total,
			mesh->n_upper_border, mesh->n_lower_border, mesh->n_upper_halo, mesh->n_lower_halo, mesh->n_neumann_edges,
			mesh->upper_border_index, mesh->lower_border_index, mesh->upper_halo_index, mesh->lower_halo_index, mesh->neumann_edges_index
	);*/

    // Read file contents
	int_t id_local = 0;
	for (int_t i=0; i<mesh->n_elements_local_total; i++) {
        int_t id, v0, v1, v2, n, n0, n1, n2;
		real_t c0, c1, c2, a;
        err = fscanf ( fp, "%ld %ld %ld %ld %lf %lf %lf %lf %ld %ld %ld %ld\n", &id, &v0, &v1, &v2, &c0, &c1, &c2, &a, &n, &n0, &n1, &n2 );
        if ( err != 12 )
        {
            fprintf ( stderr, "Error: Unable to read from file\n" );
            exit ( EXIT_FAILURE );
        }
		triangle_t *element = &mesh->elements[i];
		element->id = id;
		element->id_local = id_local;
        element->v[0] = v0;
        element->v[1] = v1;
        element->v[2] = v2;
		element->c[0] = c0;
		element->c[1] = c1;
		element->c[2] = c2;
		element->area = a;
		element->n_neighbors = n;
		element->neighbors_id[0] = n0;
		element->neighbors_id[1] = n1;
		element->neighbors_id[2] = n2;
		element->neighbors_id_local[0] = -1;
		element->neighbors_id_local[1] = -1;
		element->neighbors_id_local[2] = -1;

		id_local++;
	}

	for (int_t i=0; i<mesh->n_vertices_local; i++) {
        int_t id;
        real_t z,y,x;
        err = fscanf ( fp, "%ld %lf %lf %lf\n", &id, &x, &y, &z );
        if ( err != 4 )
        {
            fprintf ( stderr, "Error: Unable to read from file\n" );
            exit ( EXIT_FAILURE );
        }
		vertex_t *vertex = &mesh->vertices[i];
		vertex->id = id;
		vertex->x = x;
		vertex->y = y;
		vertex->z = z;
    }

	fclose(fp);
}


void
mesh_finalize ( mesh_t *mesh )
{
    free ( mesh->vertices );
    free ( mesh->elements );
}

