#define _XOPEN_SOURCE 500
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <math.h>
#include <chrono>

#include <mkl.h>

void options ( int argc, char **argv );


double
    *A, *B, *C, alpha = 1.0, beta = 0.0;
int
    m = 2048,
    n = 4000,
    k = 2048,
    max_iter = 500;

char *filename = NULL;
bool quiet = false;

void
init_dummy_data ( void )
{
    #pragma omp parallel for
    for ( int i = 0; i < (m*k); i++ )
        A[i] = (double)(i+1);
    #pragma omp parallel for
    for ( int i = 0; i < (k*n); i++ )
        B[i] = (double)(-i-1);
    #pragma omp parallel for
    for ( int i = 0; i < (m*n); i++ )
        C[i] = 0.0;
}

int
main ( int argc, char **argv )
{
    options ( argc, argv );

    A = (double *)mkl_malloc( m*k*sizeof( double ), 64 );
    B = (double *)mkl_malloc( k*n*sizeof( double ), 64 );
    C = (double *)mkl_malloc( m*n*sizeof( double ), 64 );

    /* Initialize with dummy data */
    init_dummy_data();

    // Initial call to initialize buffers
    cblas_dgemm (
        CblasRowMajor, CblasNoTrans, CblasNoTrans, 
        m, n, k, alpha, A, k, B, n, beta, C, n
    );
    double
        flop = (m*n*k + m*n*(k+2.0)),
        timings[2][max_iter];

    // Initial warmup iterations, to trigger frequency scaling
    for ( int warmup=0; warmup<max_iter/10; warmup++ )
    	cblas_dgemm (
        	CblasRowMajor, CblasNoTrans, CblasNoTrans, 
            m, n, k, alpha, A, k, B, n, beta, C, n
        );

    for ( int iter=0; iter<max_iter; iter++ )
    {
        auto t_reset_start = std::chrono::steady_clock::now();
        init_dummy_data();
        auto t_reset_end = std::chrono::steady_clock::now();
        std::chrono::duration<double> reset_diff = t_reset_end - t_reset_start;

        auto t_start = std::chrono::steady_clock::now();
        cblas_dgemm (
            CblasRowMajor, CblasNoTrans, CblasNoTrans, 
            m, n, k, alpha, A, k, B, n, beta, C, n
        );
        auto t_end = std::chrono::steady_clock::now();
        std::chrono::duration<double> diff = t_end - t_start;

        timings[0][iter] = reset_diff.count();
        timings[1][iter] = diff.count();
    }

    FILE *output = stdout;
    if ( filename != NULL )
        output = fopen ( filename, "w" );

    if ( quiet )
    {
        /* Compute the harmonic mean from all observations */
        double
            seconds_per_flop = 0.0,
            harmonic_mean,
            min = flop / (timings[1][0] * 1e9),
            max = flop / (timings[1][0] * 1e9);
        for ( int iter=0; iter<max_iter; iter++ )
        {
            seconds_per_flop += timings[1][iter]*1e9 / flop;
            if ( flop / (timings[1][iter]*1e9) > max )
                max = flop / (timings[1][iter]*1e9);
            if ( flop / (timings[1][iter]*1e9) < min )
                min = flop / (timings[1][iter]*1e9);
        }
        harmonic_mean = seconds_per_flop / max_iter;
        fprintf ( output, "%e [GFLOP/s] minimum\n", min );
        fprintf ( output, "%e [GFLOP/s] maximum\n", max );
        fprintf ( output, "%e [GFLOP/s] average\n", pow(harmonic_mean,-1.0) );
    }
    else
    {
        for ( int iter=0; iter<max_iter; iter++ )
        {
            fprintf( stderr, "%e s init\n", flop/(1e9*timings[1][iter]) );
            fprintf( output, "%e [GFLOP/s]\n", flop/(1e9*timings[1][iter]) );
        }
    }
    fclose ( output );

    mkl_free ( A );
    mkl_free ( B );
    mkl_free ( C );
    free ( filename );

    exit ( EXIT_SUCCESS );
}


const char *usage =
    "Time (m x k) matrix by (k x n) matrix multiplication with Intel MKL.\n" 
    "\t-m <int>     Set m dimension (default 2048)\n"
    "\t-k <int>     Set k dimension (default 2048)\n"
    "\t-n <int>     Set n dimension (default 4000)\n"
    "\t-o <string>  Direct output to file (default stdout)\n"
    "\t-i <int>     Set number of iterations to test (default 500)\n"
    "\t-q           Only report average result\n";


void
options ( int argc, char **argv )
{
    int o;
    while ( (o = getopt(argc,argv,"hqi:k:m:n:o:")) != -1 )
    switch ( o )
    {
        case 'k': k = strtol ( optarg, NULL, 10 ); break;
        case 'm': m = strtol ( optarg, NULL, 10 ); break;
        case 'n': n = strtol ( optarg, NULL, 10 ); break;
        case 'o': filename = strdup ( optarg ); break;
        case 'i': max_iter = strtol ( optarg, NULL, 10 ); break;
        case 'q': quiet = true; break;
        case 'h':
            fprintf ( stderr, "%s\n%s", argv[0], usage );
            exit ( EXIT_FAILURE );
            break;
    }
}
